/*
 * bfifo.c
 * Copyright (c) 2001, 2002, 2003 Corrado Santoro (csanto@diit.unict.it)
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: bfifo.c,v 1.1.1.1 2005/05/15 09:50:01 corrado Exp $
 */

#include <kernel/ntypes.h>
#include <kernel/asm.h>
#include <kernel/kprintf.h>
#include <kernel/string.h>
#include <kernel/mem.h>
#include <kernel/mm.h>
#include <kernel/driver.h>
#include <kernel/queue.h>
#include <kernel/file.h>
#include <kernel/task.h>
#include <kernel/errno.h>

#define BFIFO_SIZE  32 // size in blocks

typedef struct t_bfifo {
  t_block_queue * block_queue;
  char  name[MAX_FILENAME+1];
  uint32  refcount;
  struct t_bfifo * next;
  struct t_bfifo * prev;
} t_bfifo;

t_bfifo * bfifos = NULL;
t_semaphore bfifo_mutex = INIT_NEW_SEMAPHORE(1);


t_bfifo * find_bfifo(char * name)
{
  t_bfifo * p = bfifos;

  while (p != NULL)
  {
    if (!strcmp(p->name,name)) return p;
    p = p->next;
  }
  return NULL;
}


int _bfifo_open(file * f,char * name,int mode)
{
  t_block_queue * s;
  t_bfifo * fifo;

  semaphore_down(&bfifo_mutex);
  fifo = find_bfifo(name);
  if ((mode & O_CREAT) && (mode & O_WRONLY))
  {
    if (fifo != NULL)
    {
      semaphore_up(&bfifo_mutex);
      return -EEXIST; // fifo exists
    }
    // now create the fifo
    s = block_queue_new(BFIFO_SIZE);
    if (s == NULL)
    {
      semaphore_up(&bfifo_mutex);
      return -ENOMEM;
    }
    fifo = (t_bfifo *)kmalloc(sizeof(t_bfifo));
    if (fifo == NULL)
    {
      block_queue_delete(s);
      semaphore_up(&bfifo_mutex);
      return -ENOMEM;
    }
    // set up initial values
    fifo->refcount = 1;
    fifo->block_queue = s;
    strcpy(fifo->name,name);
    // link to fifo chain (at head)
    fifo->next = bfifos;
    fifo->prev = NULL;
    if (bfifos != NULL)
      bfifos->prev = fifo;
    bfifos = fifo;

    f->private_data = fifo;
    semaphore_up(&bfifo_mutex);
    return 0;
  }
  if ((!(mode & O_CREAT)) && (fifo != NULL))
  {
    // open existing fifo
    fifo->refcount++; // increment reference count
    f->private_data = fifo;
    semaphore_up(&bfifo_mutex);
    return 0;
  }
  semaphore_up(&bfifo_mutex);
  return -ENOENT;
}


int32 _bfifo_read(file * f,char * buf,int32 len)
{
  t_bfifo * fifo;

  fifo = (t_bfifo *)f->private_data;
  return block_queue_get(fifo->block_queue,buf,len);
}


int32 _bfifo_write(file * f,char * buf,int32 len)
{
  t_bfifo * fifo;

  fifo = (t_bfifo *)f->private_data;
  block_queue_put(fifo->block_queue,buf,len);
  return len;
}


int _bfifo_close(file * f)
{
  t_bfifo * fifo;

  fifo = (t_bfifo *)f->private_data;
  semaphore_down(&bfifo_mutex);
  if (--fifo->refcount > 0)
  {
    semaphore_up(&bfifo_mutex);
    return 0;
  }
  if ( (fifo->next == NULL) && (fifo->prev == NULL) )
  {
    bfifos = NULL;
  }
  else
  {
    if (fifo->prev != NULL)
      fifo->prev->next = fifo->next;
    if (fifo->next != NULL)
      fifo->next->prev = fifo->prev;
    if (fifo == bfifos)
      bfifos = fifo->next;
  }
  block_queue_delete(fifo->block_queue);
  kfree(fifo);
  f->private_data = NULL;
  semaphore_up(&bfifo_mutex);
  return 0;
}

t_file_drv_operations bfifo_ops = {
  _bfifo_open,
  _bfifo_read,
  _bfifo_write,
  _bfifo_close,
  NULL,   // ioctl
  NULL,   // lseek
  NULL    // fstat
};


void bfifo_init(void)
{
  if (!register_file_driver("bfifo",&bfifo_ops))
  {
    kprintf("Error registering bfifo driver");
    __cli();
    for (;;) ;
  }
  kprintf("Block-FIFOs ");
}
