/*
 * cfifo.c
 *
 * Copyright (c) 2001, 2002, 2003 Corrado Santoro (csanto@diit.unict.it)
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: cfifo.c,v 1.1.1.1 2005/05/15 09:50:01 corrado Exp $
 */

#include <kernel/ntypes.h>
#include <kernel/asm.h>
#include <kernel/kprintf.h>
#include <kernel/string.h>
#include <kernel/mem.h>
#include <kernel/mm.h>
#include <kernel/driver.h>
#include <kernel/queue.h>
#include <kernel/file.h>
#include <kernel/task.h>
#include <kernel/errno.h>

#define CFIFO_SIZE  1024

typedef struct t_cfifo {
  t_stream * stream;
  char  name[MAX_FILENAME+1];
  uint32  refcount;
  struct t_cfifo * next;
  struct t_cfifo * prev;
} t_cfifo;

t_cfifo * cfifos = NULL;
t_semaphore cfifo_mutex = INIT_NEW_SEMAPHORE(1);

t_cfifo * find_cfifo(char * name)
{
  t_cfifo * p = cfifos;
  while (p != NULL)
  {
    if (!strcmp(p->name,name)) return p;
    p = p->next;
  }
  return NULL;
}

int _cfifo_open(file * f,char * name,int mode)
{
  t_stream * s;
  t_cfifo * fifo;
  semaphore_down(&cfifo_mutex);
  fifo = find_cfifo(name);
  if ((mode & O_CREAT) && (mode & O_WRONLY))
  {
    if (fifo != NULL)
    {
      semaphore_up(&cfifo_mutex);
      return -EEXIST; // fifo exists
    }
    // now create the fifo
    s = stream_new(CFIFO_SIZE);
    if (s == NULL)
    {
      semaphore_up(&cfifo_mutex);
      return -ENOMEM;
    }
    fifo = (t_cfifo *)kmalloc(sizeof(t_cfifo));
    if (fifo == NULL)
    {
      stream_delete(s);
      semaphore_up(&cfifo_mutex);
      return -ENOMEM;
    }
    // set up initial values
    fifo->refcount = 1;
    fifo->stream = s;
    strcpy(fifo->name,name);
    // link to fifo chain (at head)
    fifo->next = cfifos;
    fifo->prev = NULL;
    if (cfifos != NULL)
      cfifos->prev = fifo;
    cfifos = fifo;

    f->private_data = fifo;
    semaphore_up(&cfifo_mutex);
    return 0;
  }
  if ((!(mode & O_CREAT)) && (fifo != NULL))
  {
    // open existing fifo
    fifo->refcount++; // increment reference count
    f->private_data = fifo;
    semaphore_up(&cfifo_mutex);
    return 0;
  }
  semaphore_up(&cfifo_mutex);
  return -ENOENT;
}

int32 _cfifo_read(file * f,char * buf,int32 len)
{
  t_cfifo * fifo;
  int count = 0;
  fifo = (t_cfifo *)f->private_data;
  while (len > 0)
  {
    buf[count] = stream_getbyte(fifo->stream);
    ++count;
    --len;
  }
  return count;
}

int32 _cfifo_write(file * f,char * buf,int32 len)
{
  t_cfifo * fifo;
  int count = 0;
  fifo = (t_cfifo *)f->private_data;
  while (len > 0)
  {
    stream_putbyte(fifo->stream,buf[count]);
    ++count;
    --len;
  }
  return count;
}

int _cfifo_close(file * f)
{
  t_cfifo * fifo;
  fifo = (t_cfifo *)f->private_data;
  semaphore_down(&cfifo_mutex);
  if (--fifo->refcount > 0)
  {
    semaphore_up(&cfifo_mutex);
    return 0;
  }
  if ( (fifo->next == NULL) && (fifo->prev == NULL) )
  {
    cfifos = NULL;
  }
  else
  {
    if (fifo->prev != NULL)
      fifo->prev->next = fifo->next;
    if (fifo->next != NULL)
      fifo->next->prev = fifo->prev;
    if (fifo == cfifos)
      cfifos = fifo->next;
  }
  stream_delete(fifo->stream);
  kfree(fifo);
  f->private_data = NULL;
  semaphore_up(&cfifo_mutex);
  return 0;
}

t_file_drv_operations cfifo_ops = {
  _cfifo_open,
  _cfifo_read,
  _cfifo_write,
  _cfifo_close,
  NULL,    // ioctl
  NULL,    // lseek
  NULL     // fstat
};

void cfifo_init(void)
{
  if (!register_file_driver("cfifo",&cfifo_ops))
  {
    kprintf("Error registering cfifo driver");
    __cli();
    for (;;) ;
  }
  kprintf("Character-FIFOs ");
}
