/*
 * console.c
 *-
 * Copyright (c) 2001, 2002, 2003 Corrado Santoro (csanto@diit.unict.it)
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: console.c,v 1.1.1.1 2005/05/15 09:50:01 corrado Exp $
 */

#include <kernel/ntypes.h>
#include <kernel/io.h>
#include <kernel/console.h>
#include <kernel/asm.h>
#include <kernel/interrupt.h>
#include <kernel/mem.h>
#include <kernel/driver.h>
#include <kernel/kprintf.h>
#include <kernel/wait.h>
#include <kernel/signals.h>

short console_x,console_y;
char console_attr = 0x7;

short console_rows = 25;
short console_cols = 80;

#define	VGA_MEM	0xb8000

void console_clear(void)
{
  kmemsetw((uint16 *)VGA_MEM,0x0720,console_cols*console_rows);
}

void console_cursor_off (void)
{
  outb (0xa, 0x3d4);
  outb (0x20, 0x3d5); // <--- bit 5 ==> cursor off
}

void console_move_cursor(void)
{
  uint16 cpos = (console_cols*console_y+console_x); //*2;

  outb(0xe,0x3d4);    // upper 8 bit of cursor address
  outb(cpos >> 8,0x3d5);

  outb(0xf,0x3d4);    // lower 8 bit of cursor address
  outb(cpos & 0xff,0x3d5);
}

void console_xor_cursor(void)
{
  uint16 cpos = console_cols*console_y+console_x;

  ((uint16*)VGA_MEM)[cpos] ^= 0x7f00;
}

void console_clear_home(void)
{
  console_x = 0;
  console_y = 0;
  console_clear();
  console_move_cursor();
}

void console_load_cursor(void)
{
  uint8 cpos_l,cpos_h;
  uint16  cpos;

  outb(0xe,0x3d4);    // upper 8 bit of cursor address
  cpos_h = inb(0x3d5);

  outb(0xf,0x3d4);    // lower 8 bit of cursor address
  cpos_l = inb(0x3d5);
 
  cpos = cpos_l | (cpos_h << 8);
  console_y = cpos / console_cols;
  console_x = cpos % console_cols;
}

void console_memput(int x,int y,char c,char attr)
{
  uint8 * vid = (uint8 *)(VGA_MEM + (console_cols*y+x)*2);
  *vid = c;
  *(vid+1) = attr;
}

void console_scroll_up(void)
{
  int i;
  kmemcpy((char*)VGA_MEM,(char*)(VGA_MEM+160),console_cols*(console_rows-1)*2);
  for (i = 0;i < console_cols;i++)
    console_memput(i,(console_rows - 1),' ',console_attr);
}

void console_write_char(char c)
{
  unsigned long flags;
  __save_flags_cli(flags);
  console_xor_cursor ();
  if (c < ' ')
  {
    switch (c)
    {
      case 0x0a:
        console_x = 0;
        if (++console_y >= console_rows)
        {
          console_scroll_up();
          --console_y;
        }
        break;
      case 0x0d:
        console_x = 0;
        break;
      case 0x08:
        if (--console_x < 0)
        {
          console_x = console_cols - 1;
          if (--console_y < 0)
            console_y = 0;
        }
        console_memput(console_x,console_y,' ',console_attr);
        break;
      case 0x0c:
        console_clear_home();
        break;
    }
  }
  else
  {
    console_memput(console_x,console_y,c,console_attr);
    if (++console_x >= console_cols)
    {
      console_x = 0;
      if (++console_y >= console_rows)
      {
        console_scroll_up();
        --console_y;
      }
    }
  }
  console_move_cursor();
  console_xor_cursor ();
  __restore_flags(flags);
}

void console_write(char * buf,int len)
{
  while (len > 0)
  {
    console_write_char(*buf);
    ++buf;
    --len;
  }
}

void console_write_string(char * buf)
{
  while (*buf != 0)
  {
    console_write_char(*buf);
    ++buf;
  }
}

void console_base_init(void)
{
  console_load_cursor();
  console_cursor_off ();
  console_xor_cursor ();
}

/*
 * Console device driver routines
 */

#define VK_LEFT_SHIFT   42
#define VK_RIGHT_SHIFT  54
#define VK_CAPSLOCK     58
#define VK_LEFT_CONTROL 29

int normal_keycodes[127] = {
  -1,0x1b,'1','2','3','4','5','6','7','8','9','0','-','=',8,
  9,'q','w','e','r','t','y','u','i','o','p','[',']',0xd,
  -1,'a','s','d','f','g','h','j','k','l',';','\'','`',-1,
  -1,'z','x','c','v','b','n','m',',','.','/',-1,-1,
  -1,' ',
};

int control_keycodes[127] = {
  -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
  -1,0x11,0x17,0x05,0x12,0x14,0x19,0x15,0x09,0x0f,0x10,'[',']',-1,
  -1,0x01,0x13,0x04,0x06,0x07,0x08,0x0a,0x0b,0x0c,-1,-1,'`',-1,
  -1,0x1a,0x18,0x03,0x16,0x02,0x0e,0x0d,-1,-1,-1,-1,-1,
  -1,-1,
};

int caps_keycodes[127] = {
  -1,0x1b,'1','2','3','4','5','6','7','8','9','0','-','=',8,
  9,'Q','W','E','R','T','Y','U','I','O','P','[',']',0xd,
  -1,'A','S','D','F','G','H','J','K','L',';','\'','`',-1,
  -1,'Z','X','C','V','B','N','M',',','.','/',-1,-1,
  -1,' ',
};

int shifted_keycodes[127] = {
  -1,0x1b,'!','@','@','$','%','^','&','*','(',')','_','+',8,
  9,'Q','W','E','R','T','Y','U','I','O','P','{','}',0xa,
  -1,'A','S','D','F','G','H','J','K','L',':','"','~',-1,
  -1,'Z','X','C','V','B','N','M','<','>','?',-1,-1,
  -1,' ',
};

#define KEYBOARD_BUFFER_LEN 128

struct {
  int num,head,tail,rnum;
  uint8 data[KEYBOARD_BUFFER_LEN];
} keyboard_buffer = { 0,0,0,0,};

#define LINE_DISCIPLINE_ECHO  0x0001
#define LINE_DISCIPLINE_RAW   0x0002
#define LINE_DISCIPLINE_CR    0x0004
#define LINE_DISCIPLINE_XOFF  0x0008


#define CHAR_XON    0x11
#define CHAR_XOFF   0x13


struct {
  uint8 shift_state;
  uint8 control_state;
  uint8 capslock_state;
  uint8 left_alt_state;
  uint8 right_alt_state;
  uint8 xoff;
  uint8 line_discipline;
} console_status =
    {0, 0, 0, 0, 0, 0,
     LINE_DISCIPLINE_ECHO | LINE_DISCIPLINE_CR | LINE_DISCIPLINE_XOFF};

wait_chan keyboard_wchan = NULL, xoff_wchan = NULL;

void line_discipline_write_char(uint8 c)
{
  if (console_status.line_discipline & LINE_DISCIPLINE_RAW)
    console_write_char(c);
  else
  {
    if ((console_status.line_discipline & LINE_DISCIPLINE_CR) && (c == 10))
    {
      console_write_char(0xd);
    }
    console_write_char(c);
  }
}

void keyboard_add_queue(int c)
{
  int ch;
  if (console_status.shift_state)
    ch = shifted_keycodes[c];
  else if (console_status.capslock_state)
    ch = caps_keycodes[c];
  else if (console_status.control_state)
    ch = control_keycodes[c];
  else
    ch = normal_keycodes[c];
  if (ch == -1)
  {
    kprintf ("%x\n",c);
    return ;
  }
  if (console_status.line_discipline & LINE_DISCIPLINE_XOFF)
  {
    if (ch == CHAR_XOFF)
    {
      console_status.xoff = 1;
      return ;
    }
    else if (ch == CHAR_XON)
    {
      console_status.xoff = 0;
      thread_wakeup_all (&xoff_wchan);
      return ;
    }
  }
  if (console_status.line_discipline & LINE_DISCIPLINE_ECHO)
    line_discipline_write_char(ch);
  if ((keyboard_buffer.num + keyboard_buffer.rnum) < KEYBOARD_BUFFER_LEN)
  {
    keyboard_buffer.data[keyboard_buffer.head] = ch;
    ++keyboard_buffer.num;
    keyboard_buffer.head = (keyboard_buffer.head + 1) % KEYBOARD_BUFFER_LEN;
    if (!(console_status.line_discipline & LINE_DISCIPLINE_RAW))
    {
      switch (ch)
      {
        case 8: //backspace
          if (keyboard_buffer.num > 0)
          {
            --keyboard_buffer.num;
            if (keyboard_buffer.head > 0) keyboard_buffer.head--;
            else keyboard_buffer.head = KEYBOARD_BUFFER_LEN - 1;
          }
          // two times! remove first the BS and then the deleted char
          if (keyboard_buffer.num > 0)
          {
            --keyboard_buffer.num;
            if (keyboard_buffer.head > 0) keyboard_buffer.head--;
            else keyboard_buffer.head = KEYBOARD_BUFFER_LEN - 1;
          }
          break;
        case 13: // enter
          if (console_status.line_discipline & LINE_DISCIPLINE_CR)
            console_write_char(0xa);
          keyboard_buffer.rnum = keyboard_buffer.num;
          keyboard_buffer.num = 0;
          thread_wakeup(&keyboard_wchan);
          break;
      }
    }
    else
    {
      keyboard_buffer.rnum = keyboard_buffer.num;
      thread_wakeup(&keyboard_wchan);
    }
  }
}

void keyboard_irq(void)
{
  uint8  c,keydown;
  c = inb(0x60);
  keydown = (c & 0x80) ^ 0x80;
  c &= 0x7f;
  if (keydown)
  {
    switch (c)
    {
      case VK_RIGHT_SHIFT:
      case VK_LEFT_SHIFT:
        console_status.shift_state = 1;
        break;
      case VK_CAPSLOCK:
        console_status.capslock_state = 1 - console_status.capslock_state;
        break;
      case VK_LEFT_CONTROL:
        console_status.control_state = 1;
        break;
      default:
        keyboard_add_queue(c);
        break;
    }
  }
  else
  {
    switch (c)
    {
      case VK_RIGHT_SHIFT:
      case VK_LEFT_SHIFT:
        console_status.shift_state = 0;
        break;
      case VK_LEFT_CONTROL:
        console_status.control_state = 0;
        break;
    }
  }
  outb(0x20,0x20);
}

int _console_open(file * f,char * name,int mode)
{
  return 0;
}

// FIXME! Cooked mode -> use EOF
int32 _console_read(file * f,char * buf,int32 len)
{
  uint8  data;
  int count = 0;
  while (len > 0)
  {
    while (keyboard_buffer.rnum == 0) thread_sleep(&keyboard_wchan);
    data = keyboard_buffer.data[keyboard_buffer.tail];
    keyboard_buffer.tail = (keyboard_buffer.tail + 1) % KEYBOARD_BUFFER_LEN;
    --keyboard_buffer.rnum;
    // Race conditions?
    // No! inc & dec should be made with a single instruction
    *buf++ = data;
    ++count;
    --len;
  }
  return count;
}

int32 _console_write(file * f,char * buf,int32 len)
{
  int saved_len = len;
  while (len > 0)
  {
    while (console_status.xoff)
      thread_sleep (&xoff_wchan);
    line_discipline_write_char(*buf++);
    len--;
  }
  return saved_len;
}

int _console_close(file * f)
{
  return 0;
}

int _console_ioctl(file *f,uint32 cmd,void * arg)
{
  kprintf("ioctl command %d\r\n",cmd);
  return 0;
}

t_file_drv_operations console_ops = {
  _console_open,
  _console_read,
  _console_write,
  _console_close,
  _console_ioctl,
  NULL, // lseek
  NULL  // fstat
};

void console_init(void)
{
  if (!register_file_driver("con",&console_ops))
  {
    kprintf("Error registering console driver");
    __cli();
    for (;;) ;
  }
  register_irq(1,keyboard_irq);
  kprintf("Console ");
}
