/*
 * file.c
 *-
 * Copyright (c) 2001, 2002, 2003 Corrado Santoro (csanto@diit.unict.it)
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: file.c,v 1.1.1.1 2005/05/15 09:50:01 corrado Exp $
 */

#include <kernel/driver.h>
#include <kernel/file.h>
#include <kernel/task.h>
#include <kernel/mem.h>
#include <kernel/mm.h>
#include <kernel/string.h>
#include <kernel/errno.h>
#include <kernel/checknull.h>

// CHECK_XXX Macros are defined in kernel/checknull.h

int parse_filename(char * name,char * drv,char * fname)
{
  char * p;
  p = strchr(name,':');
  if (p == NULL) return 0;
  strncpy(drv,name,p - name);
  drv[p - name] = 0;
  strcpy(fname,p+1);
  return 1;
}

int sys_open(char * name,int mode)
{
  t_file_drv_entry * driver;
  file  * f;
  char  drv_name[MAX_DRV_NAME+1];
  char  file_name[MAX_FILENAME+1];
  int   return_value;

  // check the correctness of file name
  if (!parse_filename(name,drv_name,file_name))
  {
      errno = EINVAL;
      return -errno;
  }
  CHECK_PROC_ERRNO_RETURN (!find_file_driver(drv_name,&driver), ENOENT);
  // allocate a new file structure
  f = (file *)kmalloc(sizeof(file));
  CHECK_NULL_ERRNO_RETURN (f, ENOMEM);
  // prepare file structure
  strcpy(f->name,file_name);
  f->ref_count = 1;
  f->driver = driver;
  f->mode = mode;
  f->eof = 0;
  f->file_position = 0;
  f->isfs = 0;
  // invoke the open function of the driver
  CHECK_NULL_ERRNO_RETURN_PROC (driver->operations.open, ENOSYS, kfree(f));
  return_value = driver->operations.open(f,file_name,mode);
  if (return_value < 0)
  {
    // open returned an error
    kfree(f);
    errno = -return_value;
    return return_value;
  }
  // find a place in the task file table
  for (return_value = 0;return_value < MAX_FILES;return_value++)
  {
    if (current_task->files[return_value] == NULL)
    {
      // place found
      current_task->files[return_value] = f;
      return return_value;
    }
  }
  // no place found. return an error
  if (f->isfs) kfree (f->driver);
  kfree(f);
  errno = EMFILE;
  return -errno;
}

long sys_write(int fid,char * buf,long size)
{
  file * f;
  file_drv_write wr;
  int retval;

  f = (file *)current_task->files[fid];
  CHECK_NULL_ERRNO_RETURN (f, ENOENT);
  CHECK_PROC_ERRNO_RETURN (!(f->mode & O_WRONLY), EPERM);

  wr = ((t_file_drv_entry *)f->driver)->operations.write;
  CHECK_NULL_ERRNO_RETURN (wr, ENOSYS);

  retval = wr(f,buf,size);
  if (retval < 0)
    errno = -retval;
  else
    f->file_position += retval;
  return retval;
}

long sys_read(int fid,char * buf,long size)
{
  file * f;
  file_drv_read rd;
  long retval;

  f = (file *)current_task->files[fid];
  CHECK_NULL_ERRNO_RETURN (f, ENOENT);
  CHECK_PROC_ERRNO_RETURN (!(f->mode & O_RDONLY), EPERM);

  if (f->eof) return 0;
  rd = ((t_file_drv_entry *)f->driver)->operations.read;
  CHECK_NULL_ERRNO_RETURN (rd, ENOSYS);

  retval = rd(f,buf,size);
  if (retval < 0)
    errno = -retval;
  else
    f->file_position += retval;
  if (retval == 0)
    f->eof = 1;
  return retval;
}

long sys_ioctl(int fid,int cmd,void * arg)
{
  file * f;
  file_drv_ioctl io;
  int retval;

  f = (file *)current_task->files[fid];
  CHECK_NULL_ERRNO_RETURN (f, ENOENT);

  io = ((t_file_drv_entry *)f->driver)->operations.ioctl;
  CHECK_NULL_ERRNO_RETURN (io, ENOSYS);

  retval = io(f,cmd,arg);
  if (retval < 0)
    errno = -retval;
  return retval;
}

int sys_fstat(int fid, struct stat * sb)
{
  file * f;
  file_drv_fstat fstat;
  int retval;

  f = (file *)current_task->files[fid];
  CHECK_NULL_ERRNO_RETURN (f, ENOENT);

  fstat = ((t_file_drv_entry *)f->driver)->operations.fstat;
  CHECK_NULL_ERRNO_RETURN (fstat, ENOSYS);

  retval = fstat (f, sb);
  if (retval < 0)
    errno = -retval;
  return retval;
}

int32 sys_lseek (int fid, int32 offset, int whence)
{
  file * f;
  file_drv_lseek lseek;
  int32 retval;

  f = (file *)current_task->files[fid];
  CHECK_NULL_ERRNO_RETURN (f, ENOENT);

  lseek = ((t_file_drv_entry *)f->driver)->operations.lseek;
  CHECK_NULL_ERRNO_RETURN (lseek, ENOSYS);

  retval = lseek (f, offset, whence);
  if (retval < 0)
    errno = -retval;
  return retval;
}

int sys_close(int fid)
{
  file * f;
  file_drv_close cl;
  int retval;

  f = (file *)current_task->files[fid];
  CHECK_NULL_ERRNO_RETURN (f, ENOENT);

  f->ref_count--;
  // ref count is greater than zero ==> do not really close
  if (f->ref_count > 0)
  {
    current_task->files[fid] = NULL;
    return 0;
  }

  cl = ((t_file_drv_entry *)f->driver)->operations.close;
  CHECK_NULL_ERRNO_RETURN (cl, ENOSYS);

  retval = cl(f);
  if (retval < 0)
  {
    errno = -retval;
    return retval;
  }
  current_task->files[fid] = NULL;
  kfree(f);
  return 0;
}

/*
DIR * sys_opendir(char * name)
{
  char  drv_name[MAX_DRV_NAME+1];
  char  file_name[MAX_FILENAME+1];
  t_mount_entry * m_e;
  DIR * d;

  // check the correctness of file name
  CHECK_PROC_ERRNO_RVAL
        (!parse_filename(name,drv_name,file_name), EINVAL, NULL);
  // find if driver is relevant to a file system (is it mounted?)
  m_e = mount_find (drv_name);
  CHECK_NULL_ERRNO_RVAL (m_e, ENOENT, NULL);

  d = (DIR *) kmalloc (sizeof (DIR));
  CHECK_NULL_ERRNO_RVAL (d, ENOMEM, NULL);

  errno = fs_open_dir (drv_name, file_name, d);
  if (errno < 0)
  {
    kfree (d);
    d = NULL;
  }
  return d;
}
*/
