/*
 * hash.c
 *-
 * Copyright (c) 2001, 2002, 2003 Corrado Santoro (csanto@diit.unict.it)
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: hash.c,v 1.1.1.1 2005/05/15 09:50:01 corrado Exp $
 */

#include <kernel/ntypes.h>
#include <kernel/string.h>
#include <kernel/mm.h>
#include <kernel/hash.h>

// hash function for string keys
int string_hash(char * s,int size)
{
  return strlen(s) % size;
}

// hash function for integer keys
int int_hash(int32 h,int size)
{
  return h % size;
}

int __hash(void * k,int size,int type)
{
  switch (type)
  {
    case HASH_TYPE_STRING:
      return string_hash((char *)k,size);
    case HASH_TYPE_INTEGER:
    default:
      return int_hash((uint32)k,size);
  }
}

int compare(void * k1,void * k2,int type)
{
  switch (type)
  {
    case HASH_TYPE_STRING:
      return (strcmp((char *)k1,(char *)k2) == 0);
    case HASH_TYPE_INTEGER:
    default:
      return ((uint32)k1 == (uint32)k2);
  }
}

// create a new hash table with the given number of buckets and the given type
t_hash_table * new_hash(int hash_type)
{
  t_hash_table * table;
  int i;
  // first allocate the table
  table = (t_hash_table *)kmalloc(sizeof(t_hash_table));
  if (table == NULL)
    return NULL;
  // initialize the fields
  table->size = HASH_DEFAULT_SIZE;
  NEW_SEMAPHORE(table->mutex,1);
  table->hash_type = hash_type;
  for (i = 0;i < table->size;i++) table->bucket[i] = NULL;
  return table;
}

// delete a table
void delete_hash(t_hash_table * table)
{
  int i;
  t_hash_element * p,* p1;
  for (i = 0;i < table->size;i++)
  {
    p = table->bucket[i];
    // delete all elements in each bucket
    while (p != NULL)
    {
      p1 = p->next;
      kfree(p);
      p = p1;
    }
  }
  kfree(table);
}

t_hash_element * not_atomic_find_hash(t_hash_table * table,void * key)
{
  t_hash_element * p;
  p = table->bucket[__hash(key,table->size,table->hash_type)];
  while (p != NULL)
  {
    if (compare(key,p->key,table->hash_type)) return p;
    p = p->next;
  }
  return NULL;
}

void critical_enter_hash(t_hash_table * table)
{
  semaphore_down(&table->mutex);
}

void critical_exit_hash(t_hash_table * table)
{
  semaphore_up(&table->mutex);
}

t_hash_element * find_hash_element(t_hash_table * table,void * key)
{
  t_hash_element * p;
  semaphore_down(&table->mutex);
  p = not_atomic_find_hash(table,key);
  semaphore_up(&table->mutex);
  return p;
}

void * find_hash(t_hash_table * table,void * key)
{
  t_hash_element * p;
  semaphore_down(&table->mutex);
  p = not_atomic_find_hash(table,key);
  semaphore_up(&table->mutex);
  if (p == NULL) return p;
  else return p->data;
}

int find_and_delete_hash(t_hash_table * table,void * key,t_hash_element * el)
{
  t_hash_element * p,* p1;
  int h;
  semaphore_down(&table->mutex);
  h = __hash(key,table->size,table->hash_type);
  p = table->bucket[h];
  p1 = NULL;
  while (p != NULL)
  {
    if (compare(key,p->key,table->hash_type))
    {
      *el = *p;
      if (p1 == NULL)
        table->bucket[h] = p->next;
      else
        p1->next = p->next;
      kfree(p);
      semaphore_up(&table->mutex);
      return 1;
    }
    p1 = p;
    p = p->next;
  }
  semaphore_up(&table->mutex);
  return 0;
}

int add_hash(t_hash_table * table,void * key,void * data)
{
  int retval;
  semaphore_down(&table->mutex);
  if (not_atomic_find_hash(table,key) != NULL)
  {
    semaphore_up(&table->mutex);
    return 0;
  }
  retval = not_atomic_add_hash (table, key, data);
  semaphore_up(&table->mutex);
  return retval;
}

int not_atomic_add_hash (t_hash_table * table,void * key,void * data)
{
  t_hash_element * p;
  int h;
  p = (t_hash_element *)kmalloc(sizeof(t_hash_element));
  if (p == NULL)
  {
    return 0;
  }
  h = __hash(key,table->size,table->hash_type);
  p->key = key;
  p->data = data;
  p->next = table->bucket[h];
  table->bucket[h] = p;
  return 1;
}
