/*
 * mm.c
 *-
 * Copyright (c) 2001, 2002, 2003 Corrado Santoro (csanto@diit.unict.it)
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: mm.c,v 1.1.1.1 2005/05/15 09:50:01 corrado Exp $
 */

/*
 * The Kernel Memory Manager: it is based on the linked list method
 * It uses the first-fit algortithm and compacts memory each time
 * a block becomes free
 */

#include <kernel/ntypes.h>
#include <kernel/asm.h>
#include <kernel/kprintf.h>
#include <kernel/interrupt.h>
#include <kernel/task.h>

#pragma pack(1)
typedef struct {
  uint8   type;
  uint8   _dummy0, _dummy1, _dummy2;
  uint32  size;
  task_struct *owner;
  struct mem_block_hdr * next;
} mem_block_hdr;

typedef struct {
  mem_block_hdr hdr;
  uint8   data;
} mem_block;
#pragma pack()

#define BLOCK_FREE  0
#define BLOCK_BUSY  1

mem_block * mm_start;

void mm_init(uint32 start,uint32 size)
{
  mm_start = (mem_block*)start;
  mm_start->hdr.type = BLOCK_FREE;
  mm_start->hdr.size = size - sizeof(mem_block_hdr);
  mm_start->hdr.next = NULL;
}

void print_mem_space(void)
{
  mem_block * p = mm_start;
  uint32  free_space = 0l;
  //__cli();
  LOCK_SCHEDULER;
  while (p != NULL)
  {
    kprintf("Block at 0x%lx, size %ld, %s",&p->data,p->hdr.size,
        ( (p->hdr.type == BLOCK_FREE) ? "Free":"Busy"));
    if (p->hdr.type == BLOCK_FREE)
      free_space += p->hdr.size;
    else
    {
      if (p->hdr.owner == NULL) kprintf(", Owner 0");
      else kprintf(", Owner %d",p->hdr.owner->tid);
    }
    kprintf("\r\n");
    p = (mem_block *)p->hdr.next;
  }
  kprintf("Total free memory %ld bytes\r\n",free_space);
  //__sti();
  UNLOCK_SCHEDULER;
}

void mem_stat(uint32 * _busy,uint32 * _free)
{
  mem_block * p = mm_start;
  *_free = 0;
  *_busy = 0;
  LOCK_SCHEDULER;
  while (p != NULL)
  {
    if (p->hdr.type == BLOCK_FREE)
      *_free += p->hdr.size;
    else
      *_busy += p->hdr.size;
    p = (mem_block *)p->hdr.next;
  }
  UNLOCK_SCHEDULER;
}

// linked list method: first-fit algorithm
void * kmalloc(uint32 size)
{
  mem_block * p;
  LOCK_SCHEDULER;
  if (size == 0)
  {
    kprintf("PANIC! Zero-size memory request\r\n");
    for (;;);
  }
  p = mm_start;
  while (p != NULL)
  {
    if ( (p->hdr.type == BLOCK_FREE) && (p->hdr.size >= size) )
    {
      // place found!
      if (p->hdr.size == size)
      {
        // wow! exact size found!
        p->hdr.type = BLOCK_BUSY;
        p->hdr.owner = current_task;
        UNLOCK_SCHEDULER;
        return &p->data;
      }
      else
      {
        uint32 remain = p->hdr.size - size; // compute remaining free size
        if (remain > sizeof(mem_block_hdr))
        {
          // there is space for another free block header
          mem_block * next;
          // compute the position
          // of the new free block
          // to be created
          next = (mem_block *)((uint32)p + sizeof(mem_block_hdr) + size);
          next->hdr.size = p->hdr.size - size - sizeof(mem_block_hdr);
          next->hdr.type = BLOCK_FREE;
          next->hdr.next = p->hdr.next;
          p->hdr.next = (struct mem_block_hdr *)next;
          p->hdr.type = BLOCK_BUSY;
          p->hdr.owner = current_task;
          p->hdr.size = size;
          UNLOCK_SCHEDULER;
          return &p->data;
        }
        /*
        else {
          // remaining space is too little, allocate the whole block
          p->hdr.type = BLOCK_BUSY;
          UNLOCK_SCHEDULER;
          //__restore_flags(flags);
          //enable();
          return &p->data;
        }
        */
      }
    }
    p = (mem_block *)p->hdr.next;
  }
  UNLOCK_SCHEDULER;
  return NULL;  // no free space for the requested size
}

void kcompact(void)
{
  mem_block * p,*first_free;
  uint32 counted_size = 0;
  p = mm_start;
  first_free = NULL;
  while (p != NULL)
  {
    if (p->hdr.type == BLOCK_FREE)
    {
      if (first_free == NULL)
      {
        // start of free blocks chain found
        first_free = p;
        counted_size = p->hdr.size;
      }
      else
      {
        // antoher free block in the chain
        counted_size += (p->hdr.size + sizeof(mem_block_hdr));
      }
    }
    else
    {
      if (first_free != NULL)
      {
        // end of free blocks chain found
        first_free->hdr.size = counted_size;
        first_free->hdr.next = (struct mem_block_hdr *)p;
        counted_size = 0;
        first_free = NULL;
      }
    }
    p = (mem_block *)p->hdr.next;
  }
  if (first_free != NULL)
  {
    // end of free blocks chain found
    first_free->hdr.size = counted_size;
    first_free->hdr.next = NULL;
  }
}

void kfree(void * ptr)
{
  mem_block * p;

  LOCK_SCHEDULER;
  p = mm_start;
  while (p != NULL)
  {
    if ( (p->hdr.type == BLOCK_BUSY) && ((void*)&p->data == ptr) )
    {
      // block to free found!
      p->hdr.type = BLOCK_FREE;
      kcompact();
      UNLOCK_SCHEDULER;
      return ;
    }
    p = (mem_block *)p->hdr.next;
  }
  UNLOCK_SCHEDULER;
}

