/*
 * nufs.c
 *-
 * Copyright (c) 2001, 2002, 2003 Corrado Santoro (csanto@diit.unict.it)
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: nufs.c,v 1.1.1.1 2005/05/15 09:50:01 corrado Exp $
 */

#include <kernel/nufs.h>
#include <kernel/mem.h>
#include <kernel/mm.h>
#include <kernel/kprintf.h>
#include <user/fcntl.h>
#include <user/string.h>
#include <user/sys/ioctl.h>

#define MAX_INODES  4096

int __compute_sb_constants (t_nufs_mem_superblock * sb)
{
  int avail_size, array_size;

  // compute available size in the superblock
  avail_size = sb->sb->block_size - sizeof (t_nufs_disk_superblock);

  // align to uint32
  avail_size &= (uint32)0xffffffc;

  // compute array size
  array_size = avail_size / 4;

  sb->sb->inode_list_size = array_size / 2;
  sb->sb->block_list_size = array_size / 2;

  // compute number of inodes per each block
  sb->sb->inodes_per_block = sb->sb->block_size / sizeof (t_nufs_disk_inode);

  //kprintf ("sizeof superblock = %d\n", sizeof (t_nufs_disk_superblock));
  //kprintf ("avail size = %d\n", avail_size);
  //kprintf ("block_size = %d\r\n", sb->sb->block_size);
  //kprintf ("sizeof (inode) = %d\r\n", sizeof (t_nufs_disk_inode));
  //kprintf ("inode_list_size = %d\r\n", sb->sb->inode_list_size);
  //kprintf ("block_list_size = %d\r\n", sb->sb->block_list_size);
  //kprintf ("inodes_per_block = %d\r\n", sb->sb->inodes_per_block);

  return 1;
}

int __compute_sb_pointers (t_nufs_mem_superblock * sb)
{
  char * buffer;

  buffer = (char *)sb->sb;

  sb->inode_free_list = (uint32 *)&buffer [sizeof (t_nufs_disk_superblock)];
  sb->block_free_list = 
      (uint32 *)&buffer [ sizeof (t_nufs_disk_superblock) + 
                          (sb->sb->inode_list_size * 4) ];
  //kprintf ("sizeof superblock = %d\n", sizeof (t_nufs_disk_superblock));
  //kprintf ("buffer = %lx\n", buffer);
  //kprintf ("inode_free_list = %lx\n", sb->inode_free_list);
  //kprintf ("block_free_list = %lx\n", sb->block_free_list);
  return 1;
}

int inode_read (t_nufs_mem_superblock * sb, int inode_num,
                t_nufs_disk_inode * inode)
{
  int block_num;
  t_nufs_disk_inode * ptr;

  block_num = inode_num / sb->sb->inodes_per_block + 2;
  if (ioctl (sb->fd, _IO_BLK_READ (block_num), (void *)sb->buffer) < 0)
  {
    return -1;
  }
  ptr = (t_nufs_disk_inode *)sb->buffer;
  *inode = ptr [inode_num % sb->sb->inodes_per_block];
  return 0;
}

int ialloc (t_nufs_disk_inode * sb)
{
  return 0;
}

void __mknod (t_nufs_disk_inode * inode, char * name, int type)
{
  kmemset ((char *)inode, 0, sizeof (t_nufs_disk_inode));
  //strcpy (inode->name, name);
  inode->type = type;
}

int mknufs (char * device)
{
  t_nufs_mem_superblock   sb;
  t_device_info dev_info;
  // t_nufs_disk_inode root;
  char *block_buffer;
  int f, i, inode_blocks_size;

  f = open (device, O_RDWR);
  if (f < 0)
  {
    return -1;
  }
  if (ioctl (f, IO_GETDEVICEINFO, &dev_info) < 0)
  {
    return -1;
  }

  block_buffer = (char *)kmalloc (dev_info.block_size);

  sb.sb = (t_nufs_disk_superblock *) block_buffer;
  sb.sb->magic_number = NUFS_MAGIC;
  sb.sb->block_size = dev_info.block_size;
  sb.sb->max_inodes = 4096;
  sb.sb->inodes_per_block = dev_info.block_size / sizeof (t_nufs_disk_inode);
  sb.fd = f;
  sb.buffer = (char *)kmalloc (dev_info.block_size);

  __compute_sb_constants (&sb);
  __compute_sb_pointers (&sb);

  // create inode free list
  for (i = 0; i < sb.sb->inode_list_size; i++)
  {
    sb.inode_free_list [i] = i;
  }

  // inode blocks starts from block 2
  // create blocks free list
  inode_blocks_size = sb.sb->max_inodes / sb.sb->inodes_per_block + 1 + 2;
  for (i = 0; i < sb.sb->block_list_size; i++)
  {
    sb.block_free_list [i] = i + inode_blocks_size;
  }

  sb.sb->free_inodes = 4096;

  if (ioctl (f, _IO_BLK_WRITE (1), (void *)block_buffer) < 0)
  {
    return -1;
  }

  // create root entry
  // __mknod (&root, ".", NUFS_TYPE_DIRECTORY);

  kfree (sb.buffer);
  kfree (block_buffer);

  close (f);
  return 0;
}

