/*
 * queue.c
 *-
 * Copyright (c) 2001, 2002, 2003 Corrado Santoro (csanto@diit.unict.it)
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: queue.c,v 1.1.1.1 2005/05/15 09:50:01 corrado Exp $
 */

#include <kernel/queue.h>
#include <kernel/mm.h>
#include <kernel/mem.h>
#include <kernel/kprintf.h>

t_stream * stream_new(int size)
{
  t_stream * q;
  q = (t_stream *)kmalloc(sizeof(t_stream));
  if (q == NULL)
    return NULL;
  q->data = (uint8 *)kmalloc(size);
  if (q->data == NULL)
  {
    kfree(q);
    return NULL;
  }
  q->size = size;
  //kprintf("New:%d,%d\r\n",size,q->size);
  q->num = q->head = q->tail = 0;
  NEW_COND(q->empty);
  NEW_COND(q->full);
  NEW_SEMAPHORE(q->mutex,1);
  return q;
}

void stream_delete(t_stream * q)
{
  kfree(q->data);
  kfree(q);
}

void stream_putbyte(t_stream * q,char data)
{
  //kprintf("Enter:a\r\n");
  semaphore_down(&q->mutex);
  //kprintf("Enter:%d,%d\r\n",q->num,q->size);
  while (q->num == q->size) cond_wait(&q->full,&q->mutex);
  q->data[q->head] = data;
  q->num++;
  q->head = (q->head + 1) % q->size;
  //kprintf("Enter:d\r\n");
  semaphore_up(&q->mutex);
  //kprintf("Enter:e\r\n");
  cond_signal(&q->empty);
}

char stream_getbyte(t_stream * q)
{
  char  data;
  //kprintf("Leave:a\r\n");
  semaphore_down(&q->mutex);
  //kprintf("Leave:b\r\n");
  while (q->num == 0) cond_wait(&q->empty,&q->mutex);
  //kprintf("Leave:c\r\n");
  data = q->data[q->tail];
  q->num--;
  q->tail = (q->tail + 1) % q->size;
  //kprintf("Leave:d\r\n");
  semaphore_up(&q->mutex);
  //kprintf("Leave:e\r\n");
  cond_signal(&q->full);
  return data;
}

void stream_put(t_stream * q,char * buf,int size)
{
  while (size > 0)
  {
    stream_putbyte(q,*buf);
    buf++;size--;
  }
}

void stream_get(t_stream * q,char * buf,int size)
{
  while (size > 0)
  {
    *buf = stream_getbyte(q);
    buf++;size--;
  }
}


t_block_queue * block_queue_new(int size)
{
  t_block_queue * q;
  int i;
  q = (t_block_queue *)kmalloc(sizeof(t_block_queue));
  if (q == NULL)
    return NULL;
  q->data = (t_message_block *)kmalloc(sizeof(t_message_block)*size);
  if (q->data == NULL)
  {
    kfree(q);
    return NULL;
  }
  q->size = size;
  //kprintf("New:%d,%d\r\n",size,q->size);
  for (i = 0;i < size;i++)
    q->data[i].data = NULL;
  q->num = q->head = q->tail = 0;
  NEW_COND(q->empty);
  NEW_COND(q->full);
  NEW_SEMAPHORE(q->mutex,1);
  return q;
}

void block_queue_delete(t_block_queue * q)
{
  int i;
  //kprintf("block freeing\r\n");
  for (i = 0;i < q->size;i++)
    if (q->data[i].data != NULL)
      kfree(q->data[i].data);
  kfree(q->data);
  kfree(q);
}

void block_queue_put(t_block_queue * q,char * buf,int size)
{
  char * p;
  semaphore_down(&q->mutex);
  while (q->num == q->size) cond_wait(&q->full,&q->mutex);
  p = (char *)kmalloc(size);
  if (p != NULL)
  {
    kmemcpy(p,buf,size);
    q->data[q->head].size = size;
    q->data[q->head].data = p;
    q->num++;
    q->head = (q->head + 1) % q->size;
    semaphore_up(&q->mutex);
    cond_signal(&q->empty);
  }
  else
    semaphore_up(&q->mutex);
}

int block_queue_get(t_block_queue * q,char * buf,int maxsize)
{
  t_message_block * block;
  semaphore_down(&q->mutex);
  while (q->num == 0) cond_wait(&q->empty,&q->mutex);
  block = &q->data[q->tail];
  //q->data[q->tail].data = NULL;
  q->num--;
  q->tail = (q->tail + 1) % q->size;
  if (block->size < maxsize) maxsize = block->size;
  kmemcpy(buf,block->data,maxsize);
  kfree(block->data);
  cond_signal(&q->full);
  semaphore_up(&q->mutex);
  return maxsize;
}

