/*
 * ramdisk.c
 *-
 * Copyright (c) 2001, 2002, 2003 Corrado Santoro (csanto@diit.unict.it)
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: ramdisk.c,v 1.1.1.1 2005/05/15 09:50:01 corrado Exp $
 */

#include <kernel/driver.h>
#include <kernel/task.h>
#include <kernel/wait.h>
#include <kernel/mem.h>
#include <kernel/mm.h>
#include <kernel/string.h>
#include <kernel/errno.h>
#include <kernel/kprintf.h>
#include <kernel/asm.h>
#include <user/stdlib.h>
#include <user/sys/ioctl.h>

#define RAMDISK_BLOCK_SIZE  512
#define RAMDISK_DISK_SIZE   (1024l * 1024l)   // 2 MBytes for each disk
#define RAMDISK_NDISKS      4                 // 4 disks

typedef struct
{
  t_semaphore disk_mutex;
  char  disk_data [RAMDISK_DISK_SIZE];
} t_ramdisk_entry;

struct ram_disks {
  t_semaphore mutex;
  t_ramdisk_entry * disk [RAMDISK_NDISKS];
} ram_disks;

int _ramdisk_open(file * f, char * name, int mode)
{
  int disk_num;
  t_ramdisk_entry * p;
  semaphore_down (&ram_disks.mutex);
  disk_num = atoi (name);
  if ( (disk_num < 0) && (disk_num > (RAMDISK_NDISKS - 1)) )
  {
    semaphore_up (&ram_disks.mutex);
    return -ENOENT;
  }
  p = ram_disks.disk [disk_num];
  if (p == NULL)
  {
    p = (t_ramdisk_entry *)kmalloc (sizeof (t_ramdisk_entry));
    if (p == NULL)
    {
      semaphore_up (&ram_disks.mutex);
      return -ENOMEM;
    }
    NEW_SEMAPHORE (p->disk_mutex, 1);
    ram_disks.disk [disk_num] = p;
  }
  f->private_data = (void *)p;
  f->file_position = 0;
  f->stat_buffer.st_size = RAMDISK_DISK_SIZE;
  semaphore_up (&ram_disks.mutex);
  return 0;
}


int32 _ramdisk_read(file * f, char * buf, int32 len)
{
  int32 global_offset, last_offset;
  t_ramdisk_entry * disk;

  disk = (t_ramdisk_entry *)f->private_data;
  global_offset = f->file_position;
  last_offset = global_offset + len;
  if (last_offset > f->stat_buffer.st_size)
  {
    last_offset = f->stat_buffer.st_size;
  }
  len = last_offset - global_offset;
  semaphore_down (&disk->disk_mutex);
  kmemcpy (buf, &disk->disk_data [global_offset], len);
  semaphore_up (&disk->disk_mutex);
  return len;
}


int32 _ramdisk_write(file * f, char * buf, int32 len)
{
  int32 global_offset, last_offset;
  t_ramdisk_entry * disk;

  disk = (t_ramdisk_entry *)f->private_data;
  global_offset = f->file_position;
  last_offset = global_offset + len;
  if (last_offset > f->stat_buffer.st_size)
  {
    last_offset = f->stat_buffer.st_size;
  }
  len = last_offset - global_offset;
  semaphore_down (&disk->disk_mutex);
  kmemcpy (&disk->disk_data [global_offset], buf, len);
  semaphore_up (&disk->disk_mutex);
  return len;
}

int32 _ramdisk_lseek (file * f, int32 offset, int whence)
{
  int32 global_offset = 0;
  t_ramdisk_entry * disk;

  disk = (t_ramdisk_entry *)f->private_data;
  switch (whence)
  {
    case SEEK_SET: global_offset = offset; break;
    case SEEK_CUR: global_offset += offset; break;
    case SEEK_END: global_offset = f->stat_buffer.st_size + offset; break;
    default: return -EINVAL;
  }
  if (global_offset > f->stat_buffer.st_size)
  {
    global_offset = f->stat_buffer.st_size;
  }
  return global_offset;
}

int _ramdisk_ioctl (file * f, uint32 cmd, void * args)
{
  uint32  block;
  t_ramdisk_entry * disk;

  disk = (t_ramdisk_entry *)f->private_data;
  // FIXME! Add checks for block number
  if (_IS_BLK_READ (cmd))
  {
    block = BLK_NUM (cmd);
    semaphore_down (&disk->disk_mutex);
    kmemcpy (args, &disk->disk_data [block * RAMDISK_BLOCK_SIZE],
             RAMDISK_BLOCK_SIZE);
    semaphore_up (&disk->disk_mutex);
    return 0;
  }
  else if (_IS_BLK_WRITE (cmd))
  {
    block = BLK_NUM (cmd);
    semaphore_down (&disk->disk_mutex);
    kmemcpy (&disk->disk_data [block * RAMDISK_BLOCK_SIZE], args,
             RAMDISK_BLOCK_SIZE);
    semaphore_up (&disk->disk_mutex);
    return 0;
  }
  else if (cmd == IO_GETDEVICEINFO)
  {
    ((t_device_info *)args)->block_size = RAMDISK_BLOCK_SIZE;
    return 0;
  }
  else
    return -EINVAL;
}

int _ramdisk_fstat (file *f, struct stat * sb)
{
  *sb = f->stat_buffer;
  return 0;
}

int _ramdisk_close (file * f)
{
  return 0;
}

t_file_drv_operations ramdisk_ops = {
  _ramdisk_open,
  _ramdisk_read,
  _ramdisk_write,
  _ramdisk_close,
  _ramdisk_ioctl,
  _ramdisk_lseek,
  _ramdisk_fstat
};

void ramdisk_init(void)
{
  int i;
  if (!register_file_driver ("ramdisk", &ramdisk_ops))
  {
    kprintf("Error registering ramdisk driver");
    __cli();
    for (;;) ;
  }
  for (i = 0;i < RAMDISK_NDISKS; i++)
  {
    ram_disks.disk [i] = NULL;
  }
  NEW_SEMAPHORE (ram_disks.mutex, 1);
  kprintf("Ramdisk ");
}
