/*
 * romfs.c
 *-
 * Copyright (c) 2001, 2002, 2003 Corrado Santoro (csanto@diit.unict.it)
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: romfs.c,v 1.1.1.1 2005/05/15 09:50:01 corrado Exp $
 */

#include <kernel/driver.h>
#include <kernel/task.h>
#include <kernel/wait.h>
#include <kernel/mem.h>
#include <kernel/mm.h>
#include <kernel/string.h>
#include <kernel/errno.h>
#include <kernel/kprintf.h>
#include <kernel/asm.h>
#include <user/stdlib.h>
#include <user/sys/ioctl.h>

typedef struct
{
  char   filename [MAX_FILENAME+1];
  uint32 file_size;
  char   *file_data;
} t_romfs_file;

typedef struct t_romfs_entry {
  t_romfs_file file;
  struct t_romfs_entry * next;
} t_romfs_entry;

struct t_romfs {
  t_romfs_entry * first;
} romfs = { NULL};

void _add_romfs_file (char * fname, int size, char * data)
{
  t_romfs_entry * p;
  p = (t_romfs_entry *)kmalloc (sizeof (t_romfs_entry));
  if (p != NULL)
  {
    strcpy (p->file.filename, fname);
    p->file.file_size = size;
    p->file.file_data = (char *)kmalloc (size);
    if (p->file.file_data == NULL)
    {
      kfree (p);
      return ;
    }
    kmemcpy (p->file.file_data, data, size);
    p->next = romfs.first;
    romfs.first = p;
  }
}

t_romfs_file * _find_romfs_file (char * name)
{
  t_romfs_entry * p = romfs.first;
  while ( (p != NULL) && (strcmp (p->file.filename, name) != 0) )
    p = p->next;
  if (p == NULL)
    return NULL;
  else
    return &p->file;
}

int _romfs_open(file * f, char * name, int mode)
{
  t_romfs_file * p;

  p = _find_romfs_file (name);
  if (p == NULL)
    return -ENOENT;

  f->private_data = (void *)p;
  f->file_position = 0;
  f->stat_buffer.st_size = p->file_size;
  return 0;
}


int32 _romfs_read(file * f, char * buf, int32 len)
{
  int32 global_offset, last_offset;
  t_romfs_file * romfile;

  romfile = (t_romfs_file *)f->private_data;
  global_offset = f->file_position;
  last_offset = global_offset + len;
  if (last_offset > f->stat_buffer.st_size)
  {
    last_offset = f->stat_buffer.st_size;
  }
  len = last_offset - global_offset;
  kmemcpy (buf, &romfile->file_data [global_offset], len);
  return len;
}


int32 _romfs_write(file * f, char * buf, int32 len)
{
  return -EINVAL;
}

int32 _romfs_lseek (file * f, int32 offset, int whence)
{
  int32 global_offset = 0;
  t_romfs_file * romfile;

  romfile = (t_romfs_file *)f->private_data;
  global_offset = f->file_position;
  switch (whence)
  {
    case SEEK_SET: global_offset = offset; break;
    case SEEK_CUR: global_offset += offset; break;
    case SEEK_END: global_offset = f->stat_buffer.st_size + offset; break;
    default: return -EINVAL;
  }
  if (global_offset > f->stat_buffer.st_size)
  {
    global_offset = f->stat_buffer.st_size;
  }
  f->file_position = global_offset;
  return global_offset;
}

int _romfs_ioctl (file * f, uint32 cmd, void * args)
{
  return -EINVAL;
}

int _romfs_fstat (file *f, struct stat * sb)
{
  *sb = f->stat_buffer;
  return 0;
}

int _romfs_close (file * f)
{
  return 0;
}

t_file_drv_operations romfs_ops = {
  _romfs_open,
  _romfs_read,
  _romfs_write,
  _romfs_close,
  _romfs_ioctl,
  _romfs_lseek,
  _romfs_fstat
};

#define ADD_ROM_FILE(name,data,size)  _add_romfs_file(name,size,data);


void romfs_init(void)
{
  if (!register_file_driver ("romfs", &romfs_ops))
  {
    kprintf("Error registering romfs driver");
    __cli();
    for (;;) ;
  }
  kprintf("Romfs ");
  #include "romfiles.inc"
}
