/*
 * rpc.c
 *-
 * Copyright (c) 2001, 2002, 2003 Corrado Santoro (csanto@diit.unict.it)
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: rpc.c,v 1.1.1.1 2005/05/15 09:50:01 corrado Exp $
 */

#include <kernel/mm.h>
#include <kernel/mem.h>
#include <kernel/task.h>
#include <kernel/rpc.h>
#include <kernel/kprintf.h>
#include <kernel/errno.h>
#include <user/string.h>
#include <user/pthread.h>
#include <user/stdio.h>
#include <user/fcntl.h>

t_rpc_svr_list * svr_list = NULL;
t_semaphore rpc_mutex = INIT_NEW_SEMAPHORE(1);

t_rpc_server * find_server (char * name)
{
  t_rpc_svr_list * p;
  p = svr_list;
  while ( (p != NULL) && (strcmp (p->server.name, name) != 0) )
    p = p->next;
  return &p->server;
}

t_rpc_proc_entry * find_procedure (t_rpc_server * svr, char * name)
{
  int i;
  for (i = 0;i < svr->procedure_count; i++)
  {
    if (strcmp (name, svr->procs[i].name) == 0)
      return &svr->procs[i];
  }
  return NULL;
}

t_rpc_server * rpc_create_program(char * name,int maxproc)
{
  t_rpc_svr_list * p;
  t_rpc_server * svr = NULL;
  semaphore_down(&rpc_mutex);
  p = (t_rpc_svr_list *)kmalloc(sizeof(t_rpc_svr_list));
  if (p == NULL) goto do_exit;
  p->server.procs = (t_rpc_proc_entry *)
                        kmalloc (sizeof(t_rpc_proc_entry)*maxproc);
  if (p->server.procs == NULL)
  {
    kfree(p);
    svr = NULL;
    goto do_exit;
  }
  p->next = svr_list;
  svr_list = p;
  svr = &p->server;
  svr->max_procedure = maxproc;
  svr->procedure_count = 0;
  strcpy(svr->name, name);
  NEW_SEMAPHORE(svr->mutex,1);
do_exit:
  semaphore_up(&rpc_mutex);
  return svr;
}

int rpc_register_call(t_rpc_server * svr,char * procname,rpc_proc proc)
{
  int id;
  id = svr->procedure_count;
  strcpy(svr->procs[id].name, procname);
  svr->procs[id].proc = proc;
  svr->procedure_count++;
  return 0;
}

int rpc_serve (t_rpc_server * svr,t_rpc_request * r)
{
  t_rpc_proc_entry  *p;
  int result = 0;
  semaphore_down (&svr->mutex);
  p = find_procedure (svr, r->p_name);
  semaphore_up (&svr->mutex);
  if (p != NULL)
  {
    result = p->proc (r->params);
  }
  else
  {
    result = -ENOENT;
  }
  return result;
}

void rpc_end(t_rpc_request * r,int retval)
{
  r->params[RPC_RESULT] = retval;
  semaphore_up (r->sync_semaphore);
}

void * rpc_start_server(void * x)
{
  int f;
  t_rpc_server *svr;
  t_rpc_request rpc_request;
  char fifoname[MAX_FILENAME];
  svr = (t_rpc_server *) x;
  sprintf (fifoname,"bfifo:%s",svr->name);
  f = open (fifoname, O_RDWR | O_CREAT);
  if (f < 0)
  {
    printf ("Unable to create RPC FIFO\n");
    return 0;
  }
  for (;;)
  {
    int retval;
    if (read (f, (char *)&rpc_request, sizeof(rpc_request)) > 0)
    {
      if (rpc_request.cmd == RPC_CALL)
      {
        retval = rpc_serve (svr, &rpc_request);
      }
      else
        retval = -EFAULT;
      rpc_end (&rpc_request,retval);
    }
  }
  return 0;
}

void rpc_start(t_rpc_server * svr)
{
  pthread_t rpc;
  pthread_create (&rpc,NULL,rpc_start_server, (void*)svr);
}

int rpc_connect (char * name)
{
  int f;
  char fifoname[MAX_FILENAME];
  sprintf (fifoname,"bfifo:%s",name);
  f = open (fifoname, O_RDWR);
  return f;
}

int rpc_call (int pgm, char * fname,uint32 * params)
{
  int retval;
  t_rpc_request r;
  t_semaphore sync = INIT_NEW_SEMAPHORE (0);
  r.cmd = RPC_CALL;
  strcpy (r.p_name, fname);
  r.params = params;
  r.sync_semaphore = &sync;
  retval = write (pgm, (char *)&r, sizeof (t_rpc_request));
  if (retval == sizeof (t_rpc_request))
  {
    semaphore_down (&sync);
    retval = r.params[0];
  }
  else
    retval = -EFAULT;
  return retval;
}

int rpc_disconnect (int pgm)
{
  close (pgm);
  return 0;
}
