/*
 * stdio.c
 *-
 * Copyright (c) 2001, 2002, 2003 Corrado Santoro (csanto@diit.unict.it)
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: stdio.c,v 1.1.1.1 2005/05/15 09:50:01 corrado Exp $
 */

#include <user/stdarg.h>
#include <kernel/ntypes.h>
#include <kernel/conv.h>
#include <kernel/string.h>
#include <user/stdlib.h>
#include <user/ctype.h>
#include <kernel/basic_printf.h>
#include <user/stdio.h>
#include <user/fcntl.h>
#include <user/unistd.h>
#include <user/errno.h>

FILE _stdout = {0};
FILE  * stdout = &_stdout;
FILE _stdin = {1};
FILE  * stdin = &_stdin;
FILE _stderr = {2};
FILE  * stderr = &_stderr;

int fputs(const char *s,FILE * f)
{
  return write(f->fid,(char *)s,strlen((char *)s));
}

int fputc(int c,FILE * f)
{
  char  _c = c;
  return write(f->fid,&_c,1);
}

int fgetc(FILE * f)
{
  char  _c;

  if (read(f->fid,&_c,1) > 0)
    return _c;
  else
    return EOF;
}

char * fgets(char * s,int size,FILE * f)
{
  while (size > 0)
  {
    int c;
    c = fgetc(f);
    if (c == 13)
      break;
    if (c == EOF)
    {
      *s = 0;
      return NULL;
    }
    *s = c;
    s++;
    size --;
  }
  *s = 0;
  return s;
}

int vfprintf(FILE *f,char *fmt,va_list ap)
  __basic_printf(f,{},fputc,fputs)

int fprintf(FILE * f,char *fmt,...)
{
  va_list ap;
  va_start(ap,fmt);
  return vfprintf(f,fmt,ap);
}

int printf(char *fmt,...)
{
  va_list ap;
  va_start(ap,fmt);
  return vfprintf(stdout,fmt,ap);
}

void sputc(char c,char *dest)
{
  char s[2] = { c, 0 };
  strcat(dest,s);
}

void sputs(char * s,char * dest)
{
  strcat(dest,s);
}

int vsprintf(char *f,char *fmt,va_list ap)
  __basic_printf(f,{*f = 0;},sputc,sputs)

int sprintf(char * f,char *fmt,...)
{
  va_list ap;
  va_start(ap,fmt);
  return vsprintf(f,fmt,ap);
}

FILE * fopen (const char * path, const char * mode)
{
  int accmode, fildes;
  FILE  * file;

  accmode = 0;
  while (*mode != 0)
  {
    switch (*mode)
    {
      case 'w': accmode |= (O_CREAT | O_WRONLY); break;
      case 'r': accmode |= O_RDONLY; break;
      case '+': accmode |= O_RDWR; break;
      case 'a': accmode |= O_APPEND; break;
    }
    ++mode;
  }
  fildes = open (path, accmode);
  if (fildes < 0)
    return NULL;
  else
  {
    file = (FILE *)malloc (sizeof (FILE));
    if (file == NULL)
    {
      errno = ENOMEM;
      return NULL;
    }
    file->fid = fildes;
    return file;
  }
}

size_t fread (void * ptr, size_t size, size_t nmemb, FILE * stream)
{
  int memread = 0;
  while (memread < nmemb)
  {
    if (read (stream->fid, (char *)ptr, size) != size)
      break;
    ptr = (char *)ptr + size;
    ++memread;
  }
  return memread;
}

size_t fwrite (const void *ptr, size_t size, size_t nmemb, FILE *stream)
{
  int memwritten = 0;
  while (memwritten < nmemb)
  {
    if (write (stream->fid, (char *)ptr, size) != size)
      break;
    ptr = (char *)ptr + size;
    ++memwritten;
  }
  return memwritten;
}

int fclose (FILE * stream)
{
  close (stream->fid);
  free (stream);
  return 0;
}

int fseek (FILE *stream, long offset, int whence)
{
  off_t r;
  r = lseek (stream->fid, offset, whence);
  if (r < 0)
    return -1;
  else
    return 0;
}

long ftell (FILE * stream)
{
  off_t r;
  r = lseek (stream->fid, 0, SEEK_CUR);
  if (r < 0)
    return -1;
  else
    return r;
}

int fileno (FILE * stream)
{
  return stream->fid;
}
