/*
 * time.c
 *-
 * Copyright (c) 2001, 2002, 2003 Corrado Santoro (csanto@diit.unict.it)
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: time.c,v 1.1.1.1 2005/05/15 09:50:01 corrado Exp $
 */

#include <kernel/timer.h>
#include <kernel/wait.h>
#include <kernel/asm.h>
#include <kernel/mm.h>
#include <kernel/kprintf.h>
#include <user/sys/time.h>

struct {
  sw_timer * list;
  t_semaphore mutex;
} sw_timers = { NULL, INIT_NEW_SEMAPHORE(1) };

int gettimeofday(struct timeval *tv, struct timezone *tz)
{
  __cli();
  *tv = curr_time;
  __sti();
  return 0;
}


/*
 * Delay Routines (sleep and usleep).
 * These routines are based on the preliminar work of
 * my students Lucia Cacciola and Cristina Galvagno.
 *                                              --Corrado Santoro
 */

sw_timer * timer_create (unsigned long usecs)
{
  struct timeval when,now;
  sw_timer  *aux;

  // Allocate a new sw_timer entry
  aux = (sw_timer *)kmalloc(sizeof(sw_timer));
  if (aux == NULL)
  {
    kprintf("sys_usleep: no more memory for timer allocation\r\n");
    return NULL;
  }
  when.tv_usec = usecs % 1000000;
  when.tv_sec = usecs / 1000000;
  gettimeofday(&now,NULL);
  // compute wake up time
  time_add(&when,&now);
  // set the task to suspend
  aux->task = current_task;
  // set the wake up time
  aux->wakeup_time = when;
  aux->skip = 0;
  return aux;
}

void timer_insert (sw_timer * aux)
{
  struct timeval when;
  sw_timer  *p,*prev;
  // insert in timer list in-order
  p = sw_timers.list;
  prev = NULL;
  while ((p != NULL) && (time_compare(&p->wakeup_time,&when) < 0))
  {
    prev = p;
    p = p->next;
  }
  if (prev == NULL)
  {
    // insert as first element (or list empty)
    aux->next = sw_timers.list;
    sw_timers.list = aux;
  }
  else
  {
    // insert in the middle of the list
    prev->next = aux;
    aux->next = p;
  }
}

void sys_usleep(unsigned long usecs)
{
  sw_timer  *aux;
  unsigned long flags;

  aux = timer_create (usecs);
  if (aux == NULL)
    return;
  // enter critical section
  //semaphore_down(&sw_timers.mutex);
  LOCK_SCHEDULER ;
  // insert in timer list in-order
  timer_insert (aux);
  // task goes sleeping...
  current_task->state = TASK_SLEEPING;
  // exit critical section
  UNLOCK_SCHEDULER ;
  __save_flags_cli (flags);
  asm ( "int $0x20" ::); // do scheduler
  __restore_flags (flags);
}

unsigned int sys_sleep(unsigned int secs)
{
  sys_usleep((unsigned long)secs * 1000000l);
  return 0;
}


/*
 * This routine wakes up a task sleeping due to a delay,
 * i.e. a call to "sleep" or "usleep".
 * It is invoked by "scheduler()" and thus runs with
 * interrupt disabled.
 */
void check_elapsed_timers(void)
{
  sw_timer *p,*aux;
  p = sw_timers.list;
  if (p == NULL)
    return;
  // scan the list until there are elapsed timers
  while ((p != NULL) && (time_compare(&curr_time,&p->wakeup_time) >= 0))
  {
    if (!p->skip)  // 'skip' means that the timer was invalidated
    {
      if (p->wait_chan_data != NULL)
        ((wait_chan)p->wait_chan_data)->skip = 1;
        // skip the wait condition (if present)
        // FIXME! Race condition?? If wait_chan_data was destroyed???
      p->task->state = TASK_READY;
    }
    aux = p->next;
    kfree(p);
    p = aux;
  }
  sw_timers.list = p;
}
