/*
 * timer.c
 *-
 * Copyright (c) 2001, 2002, 2003 Corrado Santoro (csanto@diit.unict.it)
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: timer.c,v 1.1.1.1 2005/05/15 09:50:01 corrado Exp $
 */

#include <kernel/ntypes.h>
#include <kernel/interrupt.h>
#include <kernel/io.h>
#include <kernel/kprintf.h>
#include <kernel/asm.h>
#include <kernel/timer.h>
#include <kernel/task.h>
#include <kernel/kernel.h>
#include <kernel/i8253.h>
#include <kernel/hz.h>
#include <user/sys/time.h>

extern uint32 _int20_handler;
int HZ = 50;
unsigned long USEC;
struct timeval curr_time;
task_struct timer_task;
long last_elapsed = 0;


/*
 * OK this is the interrupt call sequence
 * task -->
 *    .... isr (start.s)
 *    .... handle_interrupt(uint32)
 *    .... tick_timer
 */
void tick_timer(void)
{
  scheduler();
  outb(0x20,0x20);
}

// updates the software current time
long do_timer(void)
{
  uint8 isr;
  long  val,incrval;
  // read the ISR register of 8259-1
  outb(0xa,0x20);   // 00001010 pattern to read ISR register
  isr = inb(0x20);
  outb(0x8,0x20);   // 00001000 reset read condition
  // test if IRQ0 is being serviced
  // FIXME! What is the value of "isr" when IRQ0 is not served??
  if (isr == 0)
  {
    // ok! increment software timer
    incrval = USEC - last_elapsed;
    if (incrval < 0) incrval = 0; //FIXME! ????
    last_elapsed = 0;
  }
  else
  {
    // do_timer is called due to a process which goes to sleep
    // read current count
    val = i8253_read(0);
    // determine the number of counts from last IRQ
    val = CLOCK_BASE/HZ - val;
    // determine the microseconds elapsed
    if (val < 2000)
    {
      val = (val * 1000000l)/CLOCK_BASE;
    }
    else
    {
      // prevent long overflow
      // FIXME! Use longmax!
      val = ((val * 1000l)/(CLOCK_BASE/1000l))*1000l;
    }
    incrval = val - last_elapsed;
    last_elapsed = val;
    // FIXME! incrval SHOULD BE ALWAYS > 0!!
    if (incrval < 0)
    {
      incrval = 0;
      last_elapsed = 0;
    }
  }
  curr_time.tv_usec += incrval;
  if (curr_time.tv_usec > 1000000l)
  {
    curr_time.tv_usec = curr_time.tv_usec % 1000000l;
    curr_time.tv_sec++;
  }
  return incrval;
}

// perform the sum result = result + operand
void time_add(struct timeval *result,struct timeval *operand)
{
  result->tv_usec += operand->tv_usec;
  if (result->tv_usec > 1000000l)
  {
    result->tv_usec = result->tv_usec % 1000000l;
    result->tv_sec++;
  }
  result->tv_sec += operand->tv_sec;
}

// perform the difference result = result - operand
void time_diff(struct timeval *result,struct timeval *operand)
{
  if (result->tv_usec < operand->tv_usec)
  {
    result->tv_usec = result->tv_usec + 1000000l - operand->tv_usec;
    result->tv_sec--;
  }
  result->tv_sec -= operand->tv_sec;
}

// perform time comparison
int time_compare(struct timeval *op1,struct timeval *op2)
{
  if (op1->tv_sec > op2->tv_sec)
    return 1;
  if (op1->tv_sec < op2->tv_sec)
    return -1;
  // ok! seconds are equal, compare usecs
  if (op1->tv_usec > op2->tv_usec)
    return 1;
  if (op1->tv_usec < op2->tv_usec)
    return -1;
  return 0;
}

// Initialize timer structures and 8253 timer/counter
void init_timer(void)
{
  __cli();
  kprintf("Initializing Timer and Scheduler...");
  curr_time.tv_sec = curr_time.tv_usec = 0;
  // Timer Initialization
  // To allow scheduler to perform task switching, we map the INT 0x20 handler
  // (IRQ0) as a task gate. In this way, the interrupt task state
  // is saved by the processor and can be restored easily.

  // Thus first prepare the timer_task TSS using the _int20_handler as
  // routine entry point
  task_prepare(&timer_task,(startup_proc)&_int20_handler,NULL);
  // Then prepare in the GDT a TSS descriptor pointing to prepared TSS.
  // It is associated to selector KERNEL_TIMER_TSS
  gdt_make_tss(KERNEL_TIMER_TSS,(uint32)&timer_task.tss,sizeof(timer_task),0);
  // Now setup the IDT entry relevant to INT 0x20 as a task gate descriptor
  // pointing to the KERNEL_TIMER_TSS selector
  setup_idt_entry_task_gate(0x20,0,KERNEL_TIMER_TSS);
  // finally register in the interrupt manager our IRQ0 handler
  register_irq(0,tick_timer);
  i8253_init();
  kprintf("OK\r\n");
}

