/*
 * dir.c
 * Copyright (C) 2001,2002 Corrado Santoro (csanto@diit.unict.it)
 *
 *   The contents of this file are subject to the GNU Public License
 *   (the "License"); you may not use this file except in compliance with
 *   the License. You may obtain a copy of the License at http://www.fsf.org.
 *
 *   Software distributed under the License is distributed on an "AS IS"
 *   basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 *   License for the specific language governing rights and limitations
 *   under the License.
 *
 *   The Original Code is NUXI code.
 *
 *   The Initial Developer of the Original Code is Corrado Santoro.
 *   Portions created by Corrado Santoro are Copyright (C) 2001,2002.
 *   All Rights Reserved.
 */

#include <kernel/driver.h>
#include <kernel/memfs.h>
#include <kernel/task.h>
#include <kernel/mem.h>
#include <kernel/mm.h>
#include <kernel/string.h>
#include <kernel/errno.h>


t_dirent * _find_marked_entry (t_dir * d)
{
  t_dirent * de;
  de = d->head;
  while (de != NULL)
  {
    if (de->mark)
      return de;
    de = de->next;
  }
  return de;
}

t_dir * dir_new (void)
{
  t_dir * d;
  d = (t_dir *) kmalloc (sizeof (t_dir));
  if (d != NULL)
  {
    d->head = NULL;
    NEW_SEMAPHORE ( (d->mutex) , 1);
  }
  return d;
}

t_dirent * dir_create (t_dir * d, char * name, int type)
{
  t_dirent * de;
  semaphore_down (&d->mutex);
  de = _find_marked_entry (d);
  if (de == NULL)
  {
    de = (t_dirent *) kmalloc ( sizeof (t_dirent));
    if (de == NULL)
    {
      semaphore_up (&d->mutex);
      return NULL;
    }
    de->next = d->head;
    d->head = de;
  }
  strcpy (de->name, name);
  de->type = type;
  de->size = 0;
  de->reference = NULL;
  de->mark = 0;
  semaphore_up (&d->mutex);
  return de;
}

t_dirent * dir_scan (t_dir * d, t_dirent * de)
{
  semaphore_down (&d->mutex);
  if (de == NULL)
    de = d->head;
  else
    de = de->next;
  while (de != NULL)
  {
    if (!de->mark)
      break;
    de = de->next;
  }
  semaphore_up (&d->mutex);
  return de;
}

t_dirent * dir_find (t_dir * d, char * name)
{
  t_dirent * de;
  semaphore_down (&d->mutex);
  de = d->head;
  while (de != NULL)
  {
    if ((!de->mark) && (!strcmp (de->name, name)))
      break;
    de = de->next;
  }
  semaphore_up (&d->mutex);
  return de;
}


void dir_mark (t_dir * d, t_dirent * de)
{
  semaphore_down (&d->mutex);
  de->mark = 1;
  semaphore_up (&d->mutex);
}


void dir_destroy (t_dir * d)
{
  t_dirent * de, *p;
  semaphore_down (&d->mutex);
  de = d->head;
  while (de != NULL)
  {
    p = de->next;
    kfree (de);
    de = p;
  }
  semaphore_up (&d->mutex);
  kfree (d);
}

t_dirent * dir_path_find (t_dir * root,char * pathname)
{
  char      * p;
  t_dir     * d;
  t_dirent  * de;
  d = root;
  // always start from root
  if (pathname[0] == '/') ++pathname;
  do
  {
    p = strchr (pathname, '/');
    if (p != NULL) *p = 0;
    // find name in current dir
    de = dir_find (d, pathname);
    if (de == NULL) return NULL;
    // name found check if dir
    d = (t_dir *)de->reference;
  }
  while (p != NULL);
  return de;
}

t_dirent * mem_open_dir (t_dir * root,file * f,char * name)
{
  t_dirent * d;
  d = dir_path_find (root,name);
  if (d != NULL)
  {
    f->dev_data = d;
  }
  return d;
}
