/*
 * sysdrv.c
 * Copyright (C) Corrado Santoro (csanto@diit.unict.it)
 *
 *   The contents of this file are subject to the GNU Public License
 *   (the "License"); you may not use this file except in compliance with
 *   the License. You may obtain a copy of the License at http://www.fsf.org.
 *
 *   Software distributed under the License is distributed on an "AS IS"
 *   basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 *   License for the specific language governing rights and limitations
 *   under the License.
 *
 *   The Original Code is NUXI code.
 *
 *   The Initial Developer of the Original Code is Corrado Santoro.
 *   Portions created by Corrado Santoro are Copyright (C) 2001,2002.
 *   All Rights Reserved.
 */

#include <kernel/ntypes.h>
#include <kernel/asm.h>
#include <kernel/kprintf.h>
#include <kernel/string.h>
#include <kernel/mem.h>
#include <kernel/mm.h>
#include <kernel/driver.h>
#include <kernel/queue.h>
#include <kernel/task.h>
#include <kernel/file.h>
#include <kernel/errno.h>
#include <kernel/profile.h>
#include <kernel/kernel.h>
#include <kernel/clist.h>
#include <user/stdio.h>

typedef struct {
  t_stream * stream;
  uint32  tid;
} t_sysinfo_thread;

void _sysdrv_task_info_thread(void * x)
{
  task_struct * t;
  char data[64],c;
  t_sysinfo_thread * sth = (t_sysinfo_thread *)x;
  char s[] = " TID  PTID  State    CPUTime\n";
  stream_put(sth->stream,s,strlen(s));
  // FIXME! this walking should be atomic.... but YOU CANNOT __cli !!
  t = task_list.begin;
  while (t != NULL)
  {
    c = '?';
    switch (t->state)
    {
      case TASK_RUNNING:  c = 'X'; break;
      case TASK_READY:    c = 'R'; break;
      case TASK_SLEEPING: c = 'S'; break;
      case TASK_ENDING:   c = 'E'; break;
    }
    sprintf(data,"%4ld  %4ld     %c %6ld.%06ld\n",
            t->tid,t->p_tid,c,t->cpu_time.tv_sec,t->cpu_time.tv_usec);
    stream_put(sth->stream,data,strlen(data));
    t = t->next_ordered;
  }
  *data = 0;
  stream_put(sth->stream,data,1);
  task_end();
}

void _sysdrv_mem_info_thread(void * x)
{
  char data[64];
  int32 kbusy,kfree;
  t_sysinfo_thread * sth = (t_sysinfo_thread *)x;
  mem_stat(&kbusy,&kfree);
  sprintf(data,"Memory: Total %ld KB, Used %ld KB, Free %ld KB\n",
      (uint32)machine_memory_size / 1024l,kbusy / 1024l,kfree / 1024l);
  stream_put(sth->stream,data,strlen(data));
  *data = 0;
  stream_put(sth->stream,data,1);
  task_end();
}

void _sysdrv_release_info_thread(void * x)
{
  char data[64];
  t_sysinfo_thread * sth = (t_sysinfo_thread *)x;
  sprintf(data,"NUXI Release %s\n",KERNEL_RELEASE);
  stream_put(sth->stream,data,strlen(data));
  *data = 0;
  stream_put(sth->stream,data,1);
  task_end();
}

int _sysdrv_open(file * f,char * name,int mode)
{
  t_stream * s;
  t_sysinfo_thread * sth;
  s = stream_new(1024);
  if (s == NULL)
    return -ENOMEM;
  sth = (t_sysinfo_thread *)kmalloc(sizeof(t_sysinfo_thread));
  if (sth == NULL)
  {
    stream_delete(s);
    return -ENOMEM;
  }
  sth->stream = s;
  f->dev_data = sth;
  if (!strcmp(name,"tasks")) task_add(&sth->tid,_sysdrv_task_info_thread,sth);
  else if (!strcmp(name,"mem")) task_add(&sth->tid,_sysdrv_mem_info_thread,sth);
  else if (!strcmp(name,"release"))
    task_add(&sth->tid,_sysdrv_release_info_thread,sth);
  else {
    stream_delete(s);
    kfree(sth);
    return -ENOENT;
  }
  return 0;
}

int32 _sysdrv_read(file * f,char * buf,int32 len)
{
  t_sysinfo_thread * sth;
  uint8  data;
  int count = 0;
  sth = (t_sysinfo_thread *)f->dev_data;
  while (len > 0)
  {
    stream_get(sth->stream,&data,1);
    if (data == 0)
    {
      f->eof = 1;
      break;
    }
    buf[count] = data;
    ++count;
    --len;
  }
  return count;
}

int32 _sysdrv_write(file * f,char * buf,int32 len)
{
  return -EPERM;
}

int _sysdrv_close(file * f)
{
  t_sysinfo_thread * sth;
  sth = (t_sysinfo_thread *)f->dev_data;
  stream_delete(sth->stream);
  kfree(sth);
  f->dev_data = NULL;
  return 0;
}

t_file_drv_operations sysdrv_ops = {
  _sysdrv_open,
  _sysdrv_read,
  _sysdrv_write,
  _sysdrv_close,
  NULL    // ioctl
};

void sysdrv_init(void)
{
  if (!register_file_driver("sys",&sysdrv_ops))
  {
    kprintf("Error registering sysdrv driver");
    __cli();
    for (;;) ;
  }
  kprintf("Sysinfo ");
}
