/*
 * file.c
 * Copyright (C) 2001,2002 Corrado Santoro (csanto@diit.unict.it)
 *
 *   The contents of this file are subject to the GNU Public License
 *   (the "License"); you may not use this file except in compliance with
 *   the License. You may obtain a copy of the License at http://www.fsf.org.
 *
 *   Software distributed under the License is distributed on an "AS IS"
 *   basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 *   License for the specific language governing rights and limitations
 *   under the License.
 *
 *   The Original Code is NUXI code.
 *
 *   The Initial Developer of the Original Code is Corrado Santoro.
 *   Portions created by Corrado Santoro are Copyright (C) 2001,2002.
 *   All Rights Reserved.
 */

#include <kernel/driver.h>
#include <kernel/file.h>
#include <kernel/task.h>
#include <kernel/mem.h>
#include <kernel/mm.h>
#include <kernel/string.h>
#include <kernel/filesystem.h>
#include <kernel/dir.h>
#include <kernel/errno.h>
#include <kernel/checknull.h>

// CHECK_XXX Macros are defined in kernel/checknull.h

int parse_filename(char * name,char * drv,char * fname)
{
  char * p;
  p = strchr(name,':');
  if (p == NULL) return 0;
  strncpy(drv,name,p - name);
  drv[p - name] = 0;
  strcpy(fname,p+1);
  return 1;
}

int sys_open(char * name,int mode)
{
  t_file_drv_entry * driver, *tmp_drv;
  file  * f;
  char  drv_name[MAX_DRV_NAME+1];
  char  file_name[MAX_FILENAME+1];
  char  is_fs = 0;
  int   return_value;
  t_mount_entry * m_e;

  // check the correctness of file name
  if (!parse_filename(name,drv_name,file_name))
  {
      errno = EINVAL;
      return -errno;
  }
  // find if driver is relevant to a file system (is it mounted?)
  m_e = mount_find (drv_name);
  if (m_e != NULL)
  {
    // the drive is related to a file system thus get the file system
    // virtual drive
    CHECK_PROC_ERRNO_RETURN
        (!find_file_driver(FILESYSTEM_DRIVER_NAME,&tmp_drv), ENOENT)

    driver = (t_file_drv_entry *)kmalloc (sizeof (t_file_drv_entry));
    CHECK_NULL_ERRNO_RETURN(driver, ENOMEM);

    *driver = *tmp_drv;
    strcpy (driver->name,drv_name);
    is_fs = 1;
  }
  if (!is_fs)
  {
    // not a file system driver (not mounted)
    // find the relevant driver
    CHECK_PROC_ERRNO_RETURN (!find_file_driver(drv_name,&driver), ENOENT);
  }
  // allocate a new file structure
  f = (file *)kmalloc(sizeof(file));
  CHECK_NULL_ERRNO_RETURN (f, ENOMEM);
  // prepare file structure
  strcpy(f->name,file_name);
  f->driver = driver;
  f->mode = mode;
  f->eof = 0;
  f->file_position = 0;
  f->isfs = is_fs;
  // invoke the open function of the driver
  CHECK_NULL_ERRNO_RETURN_PROC (driver->operations.open, ENOSYS, kfree(f));
  return_value = driver->operations.open(f,file_name,mode);
  if (return_value < 0)
  {
    // open returned an error
    kfree(f);
    errno = -return_value;
    return return_value;
  }
  // find a place in the task file table
  for (return_value = 0;return_value < MAX_FILES;return_value++)
  {
    if (current_task->files[return_value] == NULL)
    {
      // place found
      current_task->files[return_value] = f;
      return return_value;
    }
  }
  // no place found. return an error
  if (f->isfs) kfree (f->driver);
  kfree(f);
  errno = EMFILE;
  return -errno;
}

long sys_write(int fid,char * buf,long size)
{
  file * f;
  file_drv_write wr;
  int retval;

  f = (file *)current_task->files[fid];
  CHECK_NULL_ERRNO_RETURN (f, ENOENT);
  CHECK_PROC_ERRNO_RETURN (!(f->mode & O_WRONLY), EPERM);

  wr = ((t_file_drv_entry *)f->driver)->operations.write;
  CHECK_NULL_ERRNO_RETURN (wr, ENOSYS);

  retval = wr(f,buf,size);
  if (retval < 0)
    errno = -retval;
  else
    f->file_position += retval;
  return retval;
}

long sys_read(int fid,char * buf,long size)
{
  file * f;
  file_drv_read rd;
  long retval;

  f = (file *)current_task->files[fid];
  CHECK_NULL_ERRNO_RETURN (f, ENOENT);
  CHECK_PROC_ERRNO_RETURN (!(f->mode & O_RDONLY), EPERM);

  if (f->eof) return 0;
  rd = ((t_file_drv_entry *)f->driver)->operations.read;
  CHECK_NULL_ERRNO_RETURN (rd, ENOSYS);

  retval = rd(f,buf,size);
  if (retval < 0)
    errno = -retval;
  else
    f->file_position += retval;
  if (retval == 0)
    f->eof = 1;
  return retval;
}

long sys_ioctl(int fid,int cmd,void * arg)
{
  file * f;
  file_drv_ioctl io;
  int retval;

  f = (file *)current_task->files[fid];
  CHECK_NULL_ERRNO_RETURN (f, ENOENT);

  io = ((t_file_drv_entry *)f->driver)->operations.ioctl;
  CHECK_NULL_ERRNO_RETURN (io, ENOSYS);

  retval = io(f,cmd,arg);
  if (retval < 0)
    errno = -retval;
  return retval;
}

int sys_close(int fid)
{
  file * f;
  file_drv_close cl;
  int retval;

  f = (file *)current_task->files[fid];
  CHECK_NULL_ERRNO_RETURN (f, ENOENT);

  cl = ((t_file_drv_entry *)f->driver)->operations.close;
  CHECK_NULL_ERRNO_RETURN (cl, ENOSYS);

  retval = cl(f);
  if (retval < 0)
  {
    errno = -retval;
    return retval;
  }
  if (f->isfs) kfree (f->driver);
  current_task->files[fid] = NULL;
  kfree(f);
  return 0;
}

/*
DIR * sys_opendir(char * name)
{
  char  drv_name[MAX_DRV_NAME+1];
  char  file_name[MAX_FILENAME+1];
  t_mount_entry * m_e;
  DIR * d;

  // check the correctness of file name
  CHECK_PROC_ERRNO_RVAL
        (!parse_filename(name,drv_name,file_name), EINVAL, NULL);
  // find if driver is relevant to a file system (is it mounted?)
  m_e = mount_find (drv_name);
  CHECK_NULL_ERRNO_RVAL (m_e, ENOENT, NULL);

  d = (DIR *) kmalloc (sizeof (DIR));
  CHECK_NULL_ERRNO_RVAL (d, ENOMEM, NULL);

  errno = fs_open_dir (drv_name, file_name, d);
  if (errno < 0)
  {
    kfree (d);
    d = NULL;
  }
  return d;
}
*/
