/*
 * filesystem.c
 * Copyright (C) 2001,2002 Corrado Santoro (csanto@diit.unict.it)
 *
 *   The contents of this file are subject to the GNU Public License
 *   (the "License"); you may not use this file except in compliance with
 *   the License. You may obtain a copy of the License at http://www.fsf.org.
 *
 *   Software distributed under the License is distributed on an "AS IS"
 *   basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 *   License for the specific language governing rights and limitations
 *   under the License.
 *
 *   The Original Code is NUXI code.
 *
 *   The Initial Developer of the Original Code is Corrado Santoro.
 *   Portions created by Corrado Santoro are Copyright (C) 2001,2002.
 *   All Rights Reserved.
 */

#include <kernel/driver.h>
#include <kernel/task.h>
#include <kernel/wait.h>
#include <kernel/mem.h>
#include <kernel/mm.h>
#include <kernel/string.h>
#include <kernel/errno.h>
#include <kernel/filesystem.h>
#include <kernel/kprintf.h>
#include <kernel/asm.h>
#include <kernel/file.h>
#include <kernel/hash.h>
#include <kernel/dir.h>

/*
 * Routines to handle file systems management
 */

t_hash_table * filesystems;

int register_filesystem (char * name, t_filesystem_operations * p)
{
  return add_hash (filesystems, name, p);
}


t_file_entry * namei (t_mount_entry * mount_entry, char * pathname,
                      t_filesystem_operations * fs, t_file_entry **fentry)
{
  char * p;
  t_file_entry * e;
  *fentry = NULL;
  e = fs->fs_get_root (mount_entry);
  if (pathname[0] == PATH_SEPARATOR) ++ pathname;
  do {
    p = strchr (pathname, PATH_SEPARATOR);
    *fentry = e;
    e = fs->fs_get_entry (mount_entry, e, pathname);
    if (e == NULL) return NULL;
    if (p != NULL)
    {
      *p = 0;
      pathname = p + 1;
    }
  } while (p != NULL);
  return e;
}


/*
 * Routines to handle mount table
 */

t_hash_table * mount_table;

t_mount_entry * mount_find (char * fsname)
{
  return find_hash (mount_table, fsname);
}

int fs_mount (char * fsname, char * driver, char * fstype)
{
  int retval = -EEXIST;
  critical_enter_hash (mount_table);
  if (not_atomic_find_hash (mount_table, fsname) == NULL)
  {
    t_mount_entry * me = (t_mount_entry *)kmalloc (sizeof (t_mount_entry));
    if (me == NULL)
    {
      critical_exit_hash (mount_table);
      return -ENOMEM;
    }
    strcpy (me->fs_name, fsname);
    strcpy (me->driver_name, driver);
    strcpy (me->fstype_name, fstype);
    if (not_atomic_add_hash (mount_table, fsname, me) == 0)
    {
      critical_exit_hash (mount_table);
      return -ENOMEM;
    }
    retval = 0;
  }
  critical_exit_hash (mount_table);
  return retval;
}

int _fs_open(file * f,char * name,int mode)
{
  t_file_drv_entry  * driver;
  t_mount_entry     * e;
  t_filesystem_operations  * fs;
  t_file_entry      * fe, * dir;
  driver = FILE_DRIVER(f);
  e = mount_find (driver->name);
  //kprintf ("opened file '%s' of filesystem '%s'\r\n",name,e->fstype_name);
  fs = (t_filesystem_operations *)find_hash (filesystems, e->fstype_name);
  if (fs == NULL) return -ENODEV;
  fe = namei (e, name, fs, &dir);
  if (mode & O_CREAT)
  {
    if (fe != NULL) return -EEXIST;
  }
  else
  {
    if (fe == NULL) return -ENOENT;
  }
  fe->filesystem = fs;
  fe->mount_info = e;
  f->current_block_offset = 0;
  f->dev_data = fe;
  fs->fs_get_next_block (e, fe, f->current_block, 0);
  f->file_length = fe->size;
  return 0;
}

int32 _fs_read(file * f,char * buf,int32 len)
{
  t_file_entry * fe;
  int i;
  if (FEOF (f)) return 0;
  fe = (t_file_entry *) f->dev_data;
  for (i = 0; i < len;i++)
  {
    buf[i] = f->current_block[f->current_block_offset];
    f->current_block_offset ++;
    if ( (f->file_position + i) >= f->file_length)
      break;
  }
  return i;
}

int32 _fs_write(file * f,char * buf,int32 len)
{
  return 0;
}

int _fs_close(file * f)
{
  return 0;
}

int fs_open_dir(char * driver_name, char * pathname, DIR * dirstream)
{
  t_mount_entry     * e;
  t_filesystem_operations  * fs;
  t_file_entry      * fe, * dir;
  e = mount_find (driver_name);
  //kprintf ("opened file '%s' of filesystem '%s'\r\n",name,e->fstype_name);
  fs = (t_filesystem_operations *)find_hash (filesystems, e->fstype_name);
  if (fs == NULL) return -ENODEV;
  fe = namei (e, pathname, fs, &dir);
  if (fe == NULL) return -ENOENT;
  fe->filesystem = fs;
  fe->mount_info = e;
  dirstream->data = fe;
  return 0;
}

int fs_read_dir (DIR * dirstream, struct dirent * dir)
{
}

t_file_drv_operations fs_ops = {
  _fs_open,
  _fs_read,
  _fs_write,
  _fs_close,
  NULL  //_fs_ioctl
};


void fs_init(void)
{
  if (!register_file_driver (FILESYSTEM_DRIVER_NAME, &fs_ops))
  {
    kprintf("Error registering fs driver");
    __cli();
    for (;;) ;
  }
  filesystems = new_hash (HASH_TYPE_STRING);
  mount_table = new_hash (HASH_TYPE_STRING);
  kprintf("Basic-FS ");
}

